CSS = """
#main-area {
    layout: horizontal;
}

#sidebar {
    width: 30%;
    min-width: 30%;
    layout: vertical;
    padding-top: 1;
}

#exp-tree, #treatment-tree {
    height: 1fr;
}

#run-btn {
    width: 100%;
}

#output-tabs {
    width: 1fr;
}
"""
